import time

import pgzrun
from pgzero.clock import Clock
from pgzero.keyboard import Keyboard
from pgzero.screen import Screen

WIDTH = 600
HEIGHT = 700

screen: Screen
keyboard: Keyboard
clock: Clock

score = 0
lives = 3
paused = False

draw_funcs = []


def draw():
    screen.clear()
    screen.draw.text(f"{score}", (WIDTH / 2, 15), color="red", fontsize=24)

    for draw_func in draw_funcs:
        draw_func()


update_funcs = []


def update(dt):
    if paused:
        return

    for update_func in update_funcs:
        update_func(dt)


MUNCHER_START_X = WIDTH / 2
MUNCHER_START_Y = (HEIGHT / 4) * 3
muncher = Actor('muncher', (MUNCHER_START_X, MUNCHER_START_Y))
muncher.images = ['muncher', 'muncher2']
muncher.vx = 150
muncher.vy = 150

draw_funcs.append(muncher.draw)


def move_muncher(dt):
    if keyboard.left:
        muncher.x -= muncher.vx * dt
    if keyboard.right:
        muncher.x += muncher.vx * dt
    if keyboard.up:
        muncher.y -= muncher.vy * dt
    if keyboard.down:
        muncher.y += muncher.vy * dt


update_funcs.append(move_muncher)

BOUNDS_X1 = 50
BOUNDS_Y1 = 50
BOUNDS_X2 = WIDTH - BOUNDS_X1
BOUNDS_Y2 = HEIGHT - BOUNDS_Y1


def keep_in_bounds(actor, dt):
    if actor.x < BOUNDS_X1:
        actor.x = BOUNDS_X1
    elif actor.x > BOUNDS_X2:
        actor.x = BOUNDS_X2

    if actor.y < BOUNDS_Y1:
        actor.y = BOUNDS_Y1
    elif actor.y > BOUNDS_Y2:
        actor.y = BOUNDS_Y2


Actor.keep_in_bounds = keep_in_bounds

update_funcs.append(muncher.keep_in_bounds)


def animate(actor):
    if not hasattr(actor, "images"):
        return
    if not hasattr(actor, "fps"):
        actor.fps = 5
    if not hasattr(actor, "next_frame"):
        actor.next_frame = time.time_ns()
    if not hasattr(actor, "frame"):
        actor.frame = 0

    now = time.time_ns()

    if now > actor.next_frame:
        actor.frame = (actor.frame + 1) % len(actor.images)
        actor.image = actor.images[actor.frame]
        while actor.next_frame < now:
            actor.next_frame += (1_000_000_000 / actor.fps)


Actor.animate = animate

draw_funcs.append(muncher.animate)


def draw_lives():
    for i in range(lives):
        screen.blit('muncher', (5 + (37 * i), 5))


draw_funcs.append(draw_lives)

GHOST_START_X = WIDTH / 2
GHOST_START_Y = (HEIGHT / 3)
GHOST_START = (GHOST_START_X, GHOST_START_Y)

blue = Actor('ghost-blue', GHOST_START)
blue.images = ['ghost-blue', 'ghost-blue2']
blue.vx = -80
blue.vy = -80

orange = Actor('ghost-orange', GHOST_START)
orange.images = ['ghost-orange', 'ghost-orange2']
orange.vx = 260
orange.vy = 60

red = Actor('ghost-red', GHOST_START)
red.images = ['ghost-red', 'ghost-red2']
red.vx = 40
red.vy = 280

pink = Actor('ghost-pink', GHOST_START)
pink.images = ['ghost-pink', 'ghost-pink2']
pink.vx = 60
pink.vy = 60

ghosts = [blue, orange, red, pink]


def draw_ghosts():
    for ghost in ghosts:
        ghost.animate()
        ghost.draw()


draw_funcs.append(draw_ghosts)


def ghost_move(ghost, dt):
    ghost.x += ghost.vx * dt
    ghost.y += ghost.vy * dt

    if ghost.x >= BOUNDS_X2:
        ghost.x = BOUNDS_X2
        ghost.vx *= -1
    elif ghost.x <= BOUNDS_X1:
        ghost.x = BOUNDS_X1
        ghost.vx *= -1

    if ghost.y >= BOUNDS_Y2:
        ghost.y = BOUNDS_Y2
        ghost.vy *= -1
    elif ghost.y <= BOUNDS_Y1:
        ghost.y = BOUNDS_Y1
        ghost.vy *= -1


for ghost in ghosts:
    import types
    ghost.move = types.MethodType(ghost_move, ghost)


def move_ghosts(dt):
    for ghost in ghosts:
        ghost.move(dt)
        ghost.keep_in_bounds(dt)


update_funcs.append(move_ghosts)


def unpause():
    global paused
    paused = False


def reset_actors():
    for ghost in ghosts:
        ghost.x = GHOST_START_X
        ghost.y = GHOST_START_Y

    muncher.x = MUNCHER_START_X
    muncher.y = MUNCHER_START_Y

    if lives <= 0:
        exit()

    clock.schedule(unpause, 2)


def check_for_ghost_collision(dt):
    global lives, paused

    for ghost in ghosts:
        if ghost.colliderect(muncher):
            lives -= 1
            paused = True
            clock.schedule(reset_actors, 2)


update_funcs.append(check_for_ghost_collision)

walls = [
    Actor('wall', (WIDTH / 4, HEIGHT / 2)),
    Actor('wall', ((WIDTH / 4) * 3, HEIGHT / 2)),
    Actor('wall', (WIDTH / 2, HEIGHT / 5)),
    Actor('wall', (WIDTH / 2, (HEIGHT / 5) * 4)),
    Actor('wall2', (WIDTH / 5, HEIGHT / 3)),
    Actor('wall2', ((WIDTH / 5) * 4, HEIGHT / 3)),
    Actor('wall2', (WIDTH / 5, (HEIGHT / 3) * 2)),
    Actor('wall2', ((WIDTH / 5) * 4, (HEIGHT / 3) * 2)),
]


def draw_walls():
    for wall in walls:
        wall.draw()


draw_funcs.append(draw_walls)


def check_for_wall_collisions(dt):
    for wall in walls:
        if wall.colliderect(muncher):
            # Try each of the edges of the wall
            if muncher.top < wall.top and muncher.bottom > wall.top:
                muncher.bottom = wall.top

            if muncher.bottom > wall.bottom and muncher.top < wall.bottom:
                muncher.top = wall.bottom

            if muncher.left < wall.left and muncher.right > wall.left:
                muncher.right = wall.left

            if muncher.right > wall.right and muncher.left < wall.right:
                muncher.left = wall.right


update_funcs.append(check_for_wall_collisions)


def create_pellets():
    global pellets
    pellets = []
    for y in range(12):
        for x in range(11):
            pos = (50 + (50 * x), 100 + (50 * y))
            collide = False
            for wall in walls:
                if wall.collidepoint(pos):
                    collide = True

            if not collide:
                pellets.append(Actor('pellet', pos))


pellets = []
create_pellets()


def draw_pellets():
    for pellet in pellets:
        pellet.draw()


draw_funcs.append(draw_pellets)


def check_for_eaten_pellets(dt):
    global score, pellets
    before = len(pellets)
    pellets = [pellet for pellet in pellets if not pellet.colliderect(muncher)]
    after = len(pellets)
    score += (before - after) * 100


update_funcs.append(check_for_eaten_pellets)


def check_for_all_pellets_eaten(dt):
    global paused
    if len(pellets) <= 0:
        paused = True
        create_pellets()
        reset_actors()


update_funcs.append(check_for_all_pellets_eaten)

import types


def chase(ghost, dt):
    if ghost.vx < 0:
        ghost.vx *= -1

    if ghost.vy < 0:
        ghost.vy *= -1

    if ghost.x < muncher.x:
        ghost.x += ghost.vx * dt
    else:
        ghost.x -= ghost.vx * dt

    if ghost.y < muncher.y:
        ghost.y += ghost.vy * dt
    else:
        ghost.y -= ghost.vy * dt


blue.move = types.MethodType(chase, blue)

fruits = [
    Actor('apple', (WIDTH / 2, HEIGHT / 2)),
    Actor('lemon', (WIDTH / 2, HEIGHT / 2)),
    Actor('strawberry', (WIDTH / 2, HEIGHT / 2)),
]

fruit = None


def draw_fruits():
    if fruit is not None:
        fruit.draw()


draw_funcs.append(draw_fruits)


def show_fruit():
    import random
    global fruit
    fruit = fruits[random.randint(0, len(fruits) - 1)]
    clock.schedule(hide_fruit, 3)


def hide_fruit():
    global fruit
    fruit = None
    clock.schedule(show_fruit, 5)


clock.schedule(show_fruit, 5)

pgzrun.go()
